import re

import pytest
from django.contrib.auth.models import Group
from django.db import connection
from django.test.utils import CaptureQueriesContext
from django.urls import reverse
from rest_framework import status

from frontend_settings.settings import settings
from tests.factories import FlagFactory, UserFactory

url = reverse("settings")


@pytest.fixture
def user():
    return UserFactory.create()


@pytest.fixture
def flags(user):
    prefix = settings.WAFFLE_FLAG_PREFIX
    enabled = FlagFactory.create(name=f"{prefix}ENABLED", everyone=True)
    disabled = FlagFactory.create(name=f"{prefix}DISABLED", everyone=False)
    not_in_response = FlagFactory.create(name="NOT_IN_RESPONSE", everyone=True)
    only_for_admin = FlagFactory.create(name=f"{prefix}ADMIN", superusers=True)
    only_for_user = FlagFactory.create(name=f"{prefix}USER", superusers=False)
    only_for_user.users.set([user])

    return {
        "enabled": enabled,
        "disabled": disabled,
        "not_in_response": not_in_response,
        "only_for_admin": only_for_admin,
        "only_for_user": only_for_user,
    }


@pytest.mark.django_db
def test_returns_ok(client):
    response = client.get(url, format="json")
    assert response.status_code == status.HTTP_200_OK


@pytest.mark.django_db
@pytest.mark.usefixtures("flags")
def test_returns_feature_flags_for_no_user(client):
    response = client.get(url)

    assert response.data["flags"]["ADMIN"] is False
    assert response.data["flags"]["USER"] is False
    assert response.data["flags"]["ENABLED"] is True
    assert response.data["flags"]["DISABLED"] is False
    assert response.data["flags"].get("NOT_IN_RESPONSE") is None


@pytest.mark.django_db
@pytest.mark.usefixtures("flags")
def test_returns_feature_flags_for_admin(admin_client):
    response = admin_client.get(url)

    assert response.data["flags"]["ADMIN"] is True
    assert response.data["flags"]["USER"] is False
    assert response.data["flags"]["ENABLED"] is True
    assert response.data["flags"]["DISABLED"] is False
    assert response.data["flags"].get("NOT_IN_RESPONSE") is None


@pytest.mark.django_db
@pytest.mark.usefixtures("flags")
def test_returns_feature_flags_for_user(client, user):
    client.force_login(user)
    response = client.get(url)

    assert response.data["flags"]["ADMIN"] is False
    assert response.data["flags"]["USER"] is True
    assert response.data["flags"]["ENABLED"] is True
    assert response.data["flags"]["DISABLED"] is False
    assert response.data["flags"].get("NOT_IN_RESPONSE") is None


@pytest.mark.django_db
def test_return_settings(client):
    response = client.get(url)
    data = response.data["settings"]

    assert data["SETTING_AAA"] == "VALUE AAA"
    assert data["SETTING_BBB"] == "VALUE BBB"
    assert data["SETTING_CCC"] == "VALUE CCC"
    assert "NOT_FRONTEND" not in data


@pytest.mark.django_db
def test_avoids_n_plus_one_on_user_groups(client):
    prefix = settings.WAFFLE_FLAG_PREFIX
    user = UserFactory.create()
    group = Group.objects.create(name="frontend-settings-group")
    user.groups.add(group)

    for index in range(5):
        flag = FlagFactory.create(name=f"{prefix}GROUP_FLAG_{index}")
        flag.groups.add(group)

    client.force_login(user)

    with CaptureQueriesContext(connection) as captured:
        response = client.get(url)

    assert response.status_code == status.HTTP_200_OK
    assert all(response.data["flags"].values())

    user_group_queries = [
        query
        for query in captured.captured_queries
        if re.search(r"\bauth_user_groups\b", query["sql"], re.IGNORECASE)
    ]
    assert len(user_group_queries) <= 2



import factory
import waffle
from django.contrib.auth import get_user_model

FlagModel = waffle.get_waffle_flag_model()
UserModel = get_user_model()


class FlagFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = FlagModel


class UserFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = UserModel



from typing import Any, Dict

import waffle
from constance import config as constance_config
from django.http import HttpRequest

from frontend_settings.settings import settings


class _CachedGroupsQuerySet:
    def __init__(self, group_ids):
        self._group_ids = group_ids

    def values_list(self, field_name, flat=False):
        if field_name == "pk" and flat:
            return list(self._group_ids)
        raise NotImplementedError


class _CachedGroupsManager:
    def __init__(self, group_ids):
        self._group_ids = group_ids

    def all(self):
        return _CachedGroupsQuerySet(self._group_ids)


class _CachedUser:
    def __init__(self, user, group_ids):
        self._user = user
        self.groups = _CachedGroupsManager(group_ids)

    def __getattr__(self, key):
        return getattr(self._user, key)


class _CachedRequest:
    def __init__(self, request, user):
        self._request = request
        self.user = user

    def __getattr__(self, key):
        return getattr(self._request, key)


def get_flags(request: HttpRequest) -> Dict[str, bool]:
    prefix = settings.WAFFLE_FLAG_PREFIX
    request_with_cached_groups = request
    user = getattr(request, "user", None)
    if user and user.is_authenticated:
        group_ids = tuple(user.groups.all().values_list("pk", flat=True))
        request_with_cached_groups = _CachedRequest(
            request, _CachedUser(user, group_ids)
        )

    model = waffle.get_waffle_flag_model()
    flags = model.objects.filter(name__startswith=prefix).values_list("name", flat=True)
    return {
        name.replace(prefix, "", 1): waffle.flag_is_active(
            request_with_cached_groups, name
        )
        for name in flags
    }


def get_settings() -> Dict[str, Any]:
    constance_key_prefix: str = settings.CONSTANCE_KEY_PREFIX
    return {
        key.replace(constance_key_prefix, ""): getattr(constance_config, key, None)
        for key in dir(constance_config)
        if key.startswith(constance_key_prefix)
    }



from django.http import HttpRequest, HttpResponse
from rest_framework import permissions, status
from rest_framework.decorators import api_view, permission_classes
from rest_framework.response import Response

from frontend_settings.utils import get_flags, get_settings


@api_view(["GET"])
@permission_classes((permissions.AllowAny,))
def settings(request: HttpRequest) -> HttpResponse:
    data = {"flags": get_flags(request), "settings": get_settings()}
    return Response(data=data, status=status.HTTP_200_OK)



from django.conf import settings as dj_settings
from django.test.signals import setting_changed
from rest_framework.settings import APISettings as _APISettings

DEFAULTS = {
    "WAFFLE_FLAG_PREFIX": "FRONTEND_",
    "CONSTANCE_KEY_PREFIX": "FRONTEND_",
}

USER_SETTINGS = getattr(dj_settings, "FRONTEND_SETTINGS", None)
IMPORT_STRINGS = []
REMOVED_SETTINGS = []


class APISettings(_APISettings):
    WAFFLE_FLAG_PREFIX: str
    CONSTANCE_KEY_PREFIX: str

    @property
    def user_settings(self):
        if not hasattr(self, "_user_settings"):
            self._user_settings = getattr(settings, "FRONTEND_SETTINGS", {})
        return self._user_settings


settings = APISettings(USER_SETTINGS, DEFAULTS, IMPORT_STRINGS)


def reload_api_settings(*_, **kwargs):
    setting = kwargs["setting"]

    if setting == "FRONTEND_SETTINGS":
        settings.reload()


setting_changed.connect(reload_api_settings)



MIDDLEWARE = (
    "django.middleware.common.CommonMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "waffle.middleware.WaffleMiddleware",
)
DEBUG_PROPAGATE_EXCEPTIONS = True
DATABASES = {
    "default": {"ENGINE": "django.db.backends.sqlite3", "NAME": ":memory:"},
}
SECRET_KEY = "not very secret in tests"
ROOT_URLCONF = "tests.urls"
INSTALLED_APPS = (
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "rest_framework",
    "waffle",
    "constance",
    "frontend_settings",
    "tests",
)
ALLOWED_HOSTS = ("*",)
CONSTANCE_BACKEND = "constance.backends.database.DatabaseBackend"

FRONTEND_SETTINGS = {
    "WAFFLE_FLAG_PREFIX": "TEST_",
    "CONSTANCE_KEY_PREFIX": "TEST_FRONTEND_",
}

CONSTANCE_CONFIG = {
    "TEST_FRONTEND_SETTING_AAA": ("VALUE AAA", ""),
    "TEST_FRONTEND_SETTING_BBB": ("VALUE BBB", ""),
    "TEST_FRONTEND_SETTING_CCC": ("VALUE CCC", ""),
    "NOT_FRONTEND": ("NOT ON SETTINGS", ""),
}



#!/usr/bin/env python
from setuptools import (
    setup,
    find_packages,
)

extras_require = {
    'test': [
        'django-constance[database]',
        'factory_boy',
        'pytest',
        'pytest-cov',
        'pytest-django',
    ],
    'lint': [
        'black',
        'flake8',
        'isort',
    ],
    'doc': [],
    'dev': ['tox', 'setuptools-scm'],
}

extras_require['dev'] += (
    extras_require['test'] +
    extras_require['lint'] +
    extras_require['doc']
)


setup(
    name='django_frontend_settings',
    use_scm_version=True,
    setup_requires=["setuptools_scm"],
    url='https://github.com/loadsmart/django-frontend-settings',
    license='MIT',
    description='Expose feature flags and settings from django waffle and django constance in an endpoint.',
    long_description=open('README.rst', 'r', encoding='utf-8').read(),
    author='Loadsmart',
    author_email='developer@loadsmart.com',
    install_requires=[
        'django>=4.1,<5',
        'djangorestframework>=3.13,<4',
        'django-picklefield',
        'django-constance>=2,<3',
        'django-waffle>=2,<3',
    ],
    python_requires='>=3.10',
    extras_require=extras_require,
    packages=find_packages(exclude=['tests']),
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        "Framework :: Django :: 4.1",
        "Framework :: Django :: 4.2",
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
    ]
)
